clear;
close all;
clc
%% 初值设定
x_0 = 0;                    %初始x位置
y_0 = 500;                  %初始y位置
v_x_0 = 50;                 %初始x速度
v_y_0 = 0;                  %初始y速度
g = 9.8;                    % 重力加速度
k_x = 0.01;                 % 阻尼系数
k_y = 0.05;                 % 阻尼系数
sigma_a_x = 0.09;           % 状态转移不确定度方差
sigma_a_y = 0.09;           % 状态转移不确定度方差
sigma_r = 64;               % 测量不确定度方差
sigma_alpha = 0.01;         % 测量不确定度方差

t = 15;                     % 仿真时间
T = 0.1;                    % 采样周期 
len = fix(t/T);             % 仿真步数
%% 真实轨迹
X = zeros(len, 4);
X(1,:) = [x_0, v_x_0, y_0, v_y_0];
for k = 2 : len
    x_k = X(k-1,1);
    v_x_k = X(k-1,2);
    y_k = X(k-1,3);
    v_y_k = X(k-1,4);
    x_k = x_k + v_x_k * T;
    v_x_k = v_x_k - (k_x*v_x_k^2 + sqrt(sigma_a_x)*randn(1,1)) * T;
    y_k = y_k + v_y_k * T;
    v_y_k = v_y_k + (k_y*v_y_k^2 - g + sqrt(sigma_a_y)*randn(1,1)) * T;   
    X(k,:) = [x_k, v_x_k, y_k, v_y_k];
end
X_temp = X;
%% 雷达测量
Z = zeros(len, 2);
for k = 1 : len
    x_k = X(k,1);
    y_k = X(k,3);
    r = sqrt(x_k^2 + y_k^2) + sqrt(sigma_r)*randn(1,1);
    alpha = atan(x_k / y_k) * 180 / pi + sqrt(sigma_alpha)*randn(1,1);
    Z(k,:) = [r, alpha];
end
%% UKF 无迹卡尔曼滤波
n = 4;                                                      % 状态变量的数量
alpha_lambda = 1; k_lambda = 1;                             % 尺度参数，决定σ点距离均值的远近
lambda = alpha_lambda^2 * (n + k_lambda) - n;               % 用于计算σ点
chi_t_1 = zeros(2*n + 1, n);                                % σ点
chi_t = zeros(2*n + 1, n);                                  % σ点
w_m = zeros(2*n + 1, 1);                                    % 计算用于计算均值
w_m(1, 1) = lambda / (n + lambda);
w_m(2:end, 1) = 1 / (2*(n + lambda));
w_c = zeros(2*n + 1, 1);                                    % 计算用于计算方差
w_c(1,1) = lambda / (n + lambda) + (1 - alpha_lambda^2 + 2);
w_c(2:end, 1) = 1 / (2*(n + lambda));              
R_k = diag([0; sigma_a_x; 0; sigma_a_y]);                   % 状态转移误差的协方差矩阵
Q_k = diag([sigma_r, sigma_alpha]);                         % 测量函数误差的协方差矩阵
sigma_k = 10 * eye(n);                                      % 最优状态协方差矩阵
sigma_bar_k= 10 * eye(n);                                   % 预测状态协方差矩阵
mu_k = [0, 40, 400, 0]';                                    % 最优状态均值矩阵
mu_bar_k = zeros(n,1);                                      % 预测的状态均值矩阵
z_k = zeros(4,1);                                           % 观测矩阵
X_UKF_est = zeros(len,n);                                   % UKF后的状态存储
for k = 1 : len
    %% 1 状态预测
    % line: 2
    temp = sqrtm((n + lambda) * sigma_k);
    for i = 1 : 2 * n + 1
       if i == 1
           chi_t_1(i,:) =  mu_k';
       elseif i <= n+ 1
           chi_t_1(i,:) = mu_k' + real(temp(i - 1, :));
       else
           chi_t_1(i,:) = mu_k' - real(temp(i - n - 1, :));
       end
    end
    % line: 3
    chi_star = zeros(2*n + 1, n);
    chi_star(:,1) = chi_t_1(:,1) + chi_t_1(:,2)*T;
    chi_star(:,2) = chi_t_1(:,2) - (k_x*chi_t_1(:,2).^2)*T;
    chi_star(:,3) = chi_t_1(:,3) + chi_t_1(:,4)*T;
    chi_star(:,4) = chi_t_1(:,4) + (k_y*chi_t_1(:,4).^2 - g)*T;
    % line: 4
    mu_bar_k  = chi_star' * w_m;
    % line: 5
    temp  = zeros(n,n);
    for i = 1 : 2 * n + 1
        temp = temp + w_c(i, 1) * (chi_star(i,:)' - mu_bar_k) * (chi_star(i,:)' - mu_bar_k)';     
    end
    sigma_bar_k = temp  + R_k;
    %% 2 观测更新
    % line: 6
    temp = sqrtm((n + lambda) * sigma_k);
    for i = 1 : 2 * n + 1
       if i == 1
           chi_t(i,:) =  mu_bar_k';
       elseif i <= n+ 1       
           chi_t(i,:) = mu_bar_k' + real(temp(i - 1, :));
       else
           chi_t(i,:) = mu_bar_k' - real(temp(i - n - 1, :));
       end
    end
    % line: 7
    x_k = chi_t(:,1);
    y_k = chi_t(:,3);
    r = sqrt(x_k.^2 + y_k.^2);
    alpha = atan(x_k ./ y_k) * 180 / pi;
    Z_k = [r,alpha] ;
    % line: 8
    z_k = w_m' * Z_k;
    % line: 9
    temp = zeros(2,2);
    for i = 1 : 2 * n + 1
        temp = temp + w_c(i, 1) * (Z_k(i,:) - z_k)' * (Z_k(i,:) - z_k);     
    end
    S_k = temp  + Q_k;   
    % line: 10
    Sigma_x_z = zeros(4,2);
    for i = 1 : 2 * n + 1
        Sigma_x_z = Sigma_x_z + w_c(i, 1) * (chi_t(i,:)' - mu_bar_k) * (Z_k(i,:) - z_k);     
    end
    % line: 11
    K_k = Sigma_x_z * S_k^(-1);
    % line: 12
    mu_k = mu_bar_k + K_k * (Z(k,:) - z_k)';  
    % line: 13
    sigma_k = sigma_bar_k - K_k * S_k * K_k';
    %% 3 存储EKF后的值
    X_UKF_est(k,:) = mu_k;
end
%% EKF 扩展卡尔曼滤波
R_k = diag([0; sigma_a_x; 0; sigma_a_y]);	% 状态转移误差的协方差矩阵
Q_k = diag([sigma_r, sigma_alpha]);         % 测量函数误差的协方差矩阵
sigma_k = 10 * eye(4);                      % 最优状态协方差矩阵
sigma_bar_k= 10 * eye(4);                   % 预测状态协方差矩阵
mu_k = [0, 40, 400, 0]';                    % 最优状态均值矩阵
mu_bar_k = zeros(4,1);                      % 预测的状态均值矩阵
z_k = zeros(4,1);                           % 观测矩阵
X_EKF_est = zeros(len,4);                       %EKF后的状态存储
for k = 1 : len
    % 1 状态预测
    x1 = mu_k(1) + mu_k(2)*T;
    v_x1 = mu_k(2) - (k_x*mu_k(2)^2)*T;
    y1 = mu_k(3) + mu_k(4)*T;
    v_y1 = mu_k(4) + (k_y*mu_k(4)^2 - g)*T;
    mu_bar_k = [x1; v_x1; y1; v_y1];            % 预测的均值   
    G_k = zeros(4,4);                       % 状态雅可比矩阵
    G_k(1,1) = 1; G_k(1,2) = T;
    G_k(2,2) = 1 - 2*k_x*v_x1*T;
    G_k(3,3) = 1; G_k(3,4) = T;
    G_k(4,4) = 1 + 2*k_y*v_y1*T;    
    sigma_bar_k = G_k * sigma_k * G_k' + R_k;
    % 2 观测更新
    r = sqrt(x1*x1+y1*y1);
    alpha = atan(x1/y1)*180/pi;
    z_k = [r, alpha]';                      % h(\bar{\mu}_t) 
    H_k = zeros(2,4);                       % 状态雅可比矩阵
    x = mu_k(1); y = mu_k(3); 
    H_k(1,1) = x / r;   H_k(1,3) = y / r;
    H_k(2,1) = (1/y)/(1 + (x/y)^2);   H_k(2,3) = (-x/(y^2))/(1 + (x/y)^2);
    K_k = sigma_bar_k * H_k' * (H_k * sigma_bar_k * H_k' + Q_k)^(-1);
    mu_k = mu_bar_k + K_k * (Z(k,:)' - z_k);
    sigma_k = (eye(4) - K_k * H_k) * sigma_bar_k;
    % 3 存储EKF后的值
    X_EKF_est(k,:) = mu_k;
end
%% 绘图 
figure, hold on, grid on;
plot(X(:,1),X(:,3),'-g');                                               %真实位置
plot(Z(:,1).*sin(Z(:,2)*pi/180), Z(:,1).*cos(Z(:,2)*pi/180),'-b');      %观测位置
plot(X_UKF_est(:,1),X_UKF_est(:,3), 'c');                               %UKF最优位置
plot(X_EKF_est(:,1),X_EKF_est(:,3), 'r');                               %EKF最优位置
xlabel('X'); 
ylabel('Y'); 
title('UKF simulation');
legend('real', 'measurement', 'UKF', 'EKF');
axis([-5,230,290,530]);

figure, hold on, grid on;
T_plot = linspace(0, T, len);
plot(T_plot, abs(X(:,1) - X_UKF_est(:,1)),'-c');                                      %UKF X误差
plot(T_plot, abs(X(:,1) - X_EKF_est(:,1)), '-r');                                     %EKF X误差 
xlabel('t'); 
ylabel('X error'); 
title('UKF EKF X error simulation');
legend('UKF X error', 'EKF X error');

figure, hold on, grid on;
plot(T_plot, abs(X(:,3) - X_UKF_est(:,3)),'-c');                                      %UKF Y误差
plot(T_plot, abs(X(:,3) - X_EKF_est(:,3)), '-r');                                     %EKF Y误差 
xlabel('t'); 
ylabel('Y error'); 
title('UKF EKF Y error simulation');
legend('UKF Y error', 'EKF Y error');
